// ClipSpyDlg.cpp : implementation file
//

#include "stdafx.h"
#include "ClipSpy.h"
#include "ClipSpyDlg.h"
#include "About.h"
#include "FolderMenu.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define ID_SENDTOFIRST 1000
#define ID_SENDTOLAST 10000

namespace
{
	BOOL IsMemoryFormat(UINT nFormat)
	{
		switch (nFormat)
		{
		case CF_BITMAP:
		case CF_METAFILEPICT:
		case CF_PALETTE:
		case CF_ENHMETAFILE:
			return FALSE;
		}
		return TRUE;
	}

	CString GetFormatDescription(UINT nFormat)
	{
		const int nLen = 256;
		TCHAR szFormat[nLen];
		CString strName;
		switch (nFormat)
		{
		// Predefined formats
		case CF_TEXT: strName = "CF_TEXT"; break;
		case CF_BITMAP: strName = "CF_BITMAP"; break;
		case CF_METAFILEPICT: strName = "CF_METAFILEPICT"; break;
		case CF_SYLK: strName = "CF_SYLK"; break;
		case CF_DIF: strName = "CF_DIF"; break;
		case CF_TIFF: strName = "CF_TIFF"; break;
		case CF_OEMTEXT: strName = "CF_OEMTEXT"; break;
		case CF_DIB: strName = "CF_DIB"; break;
		case CF_PALETTE: strName = "CF_PALETTE"; break;
		case CF_PENDATA: strName = "CF_PENDATA"; break;
		case CF_RIFF: strName = "CF_RIFF"; break;
		case CF_WAVE: strName = "CF_WAVE"; break;
		case CF_UNICODETEXT: strName = "CF_UNICODETEXT"; break;
		case CF_ENHMETAFILE: strName = "CF_ENHMETAFILE"; break;
#if(WINVER >= 0x0400)
		case CF_HDROP: strName = "CF_HDROP"; break;
		case CF_LOCALE: strName = "CF_LOCALE"; break;
		case CF_MAX: strName = "CF_MAX"; break;
#endif /* WINVER >= 0x0400 */
		case CF_OWNERDISPLAY: strName = "CF_OWNERDISPLAY"; break;
		case CF_DSPTEXT: strName = "CF_DSPTEXT"; break;
		case CF_DSPBITMAP: strName = "CF_DSPBITMAP"; break;
		case CF_DSPMETAFILEPICT: strName = "CF_DSPMETAFILEPICT"; break;
		case CF_DSPENHMETAFILE: strName = "CF_DSPENHMETAFILE"; break;

		// Custom or unknown formats
		default:
			if (GetClipboardFormatName(nFormat, szFormat, nLen - 1))
				strName = szFormat;
			else 
				strName = TEXT("Unknown");
			break;
		}
		return strName;
	}

	CString GetMediumDescription(DWORD tymed)
	{
		CString str;
		if (tymed & TYMED_HGLOBAL) str += _T("HGLOBAL, ");
		if (tymed & TYMED_FILE) str += _T("FILE, ");
		if (tymed & TYMED_ISTREAM) str += _T("ISTREAM, ");
		if (tymed & TYMED_ISTORAGE) str += _T("ISTORAGE, ");
		if (tymed & TYMED_GDI) str += _T("GDI, ");
		if (tymed & TYMED_MFPICT) str += _T("MFPICT, ");
		if (tymed & TYMED_ENHMF) str += _T("ENHMF, ");
		if (str.GetLength() > 1) str = str.Left(str.GetLength() - 2);
		return str;
	}
}


/////////////////////////////////////////////////////////////////////////////
// CClipSpyDlg dialog

CClipSpyDlg::CClipSpyDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CClipSpyDlg::IDD, pParent),
	  CPersistence("Main"),
	  m_dropTarget(*this)
{
	//{{AFX_DATA_INIT(CClipSpyDlg)
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);

	m_fDoneInitDialog = FALSE;
	m_hwndClipChainNext = NULL;
	m_pDataObject = NULL;

	m_imagelist.Create(IDB_LISTIMAGE, 16, 1, RGB(255, 0, 255));
}

void CClipSpyDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CClipSpyDlg)
	DDX_Control(pDX, IDC_LIST1, m_wndList);
	DDX_Control(pDX, IDC_TEXT1, m_wndText1);
	DDX_Control(pDX, IDCANCEL, m_wndClose);
	DDX_Control(pDX, IDC_BUTTON1, m_wndRefresh);
	DDX_Control(pDX, IDC_CHECK1, m_wndUseOle);
	DDX_Check(pDX, IDC_CHECK1, m_fUseOleClipboard);
	//}}AFX_DATA_MAP
}

void CClipSpyDlg::OnExchangePersistentData()
{
	PDX_WindowRect("Position", this);
	PDX_Bool("AlwaysOnTop", m_fAlwaysOnTop, FALSE);
	PDX_Bool("OleClipboard", m_fUseOleClipboard, FALSE);
}

void CClipSpyDlg::SetTopMost(BOOL fTopMost)
{
	SetWindowPos(fTopMost ? &wndTopMost : &wndNoTopMost,
		0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
}

//
// Drag drop stuff
//
DROPEFFECT CClipSpyDlg::OnDragEnter(CWnd* pWnd, COleDataObject* pDataObject, DWORD dwKeyState, CPoint point)
{
	m_pDataObject = pDataObject;
	Refresh();
	return OnDragOver(pWnd, pDataObject, dwKeyState, point);
}

DROPEFFECT CClipSpyDlg::OnDragOver(CWnd* pWnd, COleDataObject* pDataObject, DWORD dwKeyState, CPoint point)
{
	return DROPEFFECT_NONE;
}

void CClipSpyDlg::OnDragLeave(CWnd* pWnd)
{
	m_pDataObject = NULL;
	Refresh();
}

BOOL CClipSpyDlg::OnDrop(CWnd* pWnd, COleDataObject* pDataObject, DROPEFFECT dropEffect, CPoint point)
{
	return FALSE;
}

BEGIN_MESSAGE_MAP(CClipSpyDlg, CDialog)
	//{{AFX_MSG_MAP(CClipSpyDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_WM_SIZE()
	ON_WM_CLOSE()
	ON_WM_DESTROY()
	ON_WM_DRAWCLIPBOARD()
	ON_WM_CHANGECBCHAIN()
	ON_WM_GETMINMAXINFO()
	ON_NOTIFY(LVN_BEGINDRAG, IDC_LIST1, OnBegindragList)
	ON_WM_CONTEXTMENU()
	ON_BN_CLICKED(IDC_BUTTON1, OnButton1)
	ON_BN_CLICKED(IDC_CHECK1, OnCheckOleClipboard)
	ON_WM_INITMENUPOPUP()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CClipSpyDlg message handlers

BOOL CClipSpyDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);
	ASSERT((IDM_ALWAYSONTOP & 0xFFF0) == IDM_ALWAYSONTOP);
	ASSERT(IDM_ALWAYSONTOP < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	CString strAboutMenu;
	strAboutMenu.LoadString(IDS_ABOUTBOX);
	if (!strAboutMenu.IsEmpty())
	{
		pSysMenu->AppendMenu(MF_SEPARATOR);
		pSysMenu->AppendMenu(MF_STRING, IDM_ALWAYSONTOP, "Always On &Top");
		pSysMenu->AppendMenu(MF_SEPARATOR);
		pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
	}

	// Disable fancy system menu for version higher than 4.0, as 
	// Windows 98 does its own. :-(
	OSVERSIONINFO info; info.dwOSVersionInfoSize = sizeof(info);
	GetVersionEx(&info);
	int nVer = 100 * info.dwMajorVersion + info.dwMinorVersion;
	if (nVer <= 400)
	{
		m_menuSystem.Register(this);
		m_menuSystem.AddToolbar(IDR_SYSTOOLBAR);
		m_menuSystem.Attach(*pSysMenu, TRUE);
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	HICON hiconSmall = (HICON)LoadImage(AfxGetInstanceHandle(), 
		MAKEINTRESOURCE(IDR_MAINFRAME), IMAGE_ICON, 
		GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON),
		LR_DEFAULTCOLOR);
	SetIcon(hiconSmall, FALSE);		// Set small icon

	// Get initial sizes
	CRect rect, rectClient;
	GetClientRect(&rectClient);

	// Use initial size as min tracking size
	GetWindowRect(&rect);
	m_ptMinSize = CPoint(rect.Width(), rect.Height());

	// Get listbox spacing
	m_wndList.GetWindowRect(&rect);
	ScreenToClient(&rect);
	m_posList = rectClient.BottomRight() - rect.BottomRight();

	// Get button spacing
	m_wndRefresh.GetWindowRect(&rect);
	ScreenToClient(&rect);
	m_posRefresh = rectClient.BottomRight() - rect.TopLeft();
	m_wndClose.GetWindowRect(&rect);
	ScreenToClient(&rect);
	m_posClose = rectClient.BottomRight() - rect.TopLeft();
	m_wndUseOle.GetWindowRect(&rect);
	ScreenToClient(&rect);
	m_posUseOle = CSize(rect.left, rectClient.bottom - rect.top);

	// Figure out min tracking size (based on text above listbox)
//	m_wndText1.GetWindowRect(&rect);
//	ScreenToClient(&rect);
//	m_ptMinSize.x = 2 * rect.left + rect.Width();

	// Add ourself to clipboard chain
	m_hwndClipChainNext = SetClipboardViewer();

	m_fDoneInitDialog = TRUE;

	// Load window position
	LoadPersistentData();
	SetTopMost(m_fAlwaysOnTop);
	UpdateData(FALSE); // set up checkbox

	// Register the drop target
	VERIFY(m_dropTarget.Register(this));

	// Fill the listbox
	Refresh();

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CClipSpyDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	switch (nID & 0xFFF0)
	{
	case IDM_ABOUTBOX:
		{
			CAboutDlg dlgAbout;
			dlgAbout.DoModal();
		}
		break;

	case IDM_ALWAYSONTOP:
		SetTopMost(m_fAlwaysOnTop = !m_fAlwaysOnTop);
		break;

	default:
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CClipSpyDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CClipSpyDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

void CClipSpyDlg::OnSize(UINT nType, int cx, int cy) 
{
	CDialog::OnSize(nType, cx, cy);
	
	if (!m_fDoneInitDialog)
		return;
	
	// Move controls
	CRect rect, rectClient(0, 0, cx, cy);

	// Move listbox
	m_wndList.GetWindowRect(&rect);
	ScreenToClient(&rect);
	rect.right = rectClient.right - m_posList.cx;
	rect.bottom = rectClient.bottom - m_posList.cy;
	m_wndList.MoveWindow(rect);
	m_wndList.Invalidate();

	// Move 'Close' button
	m_wndClose.GetWindowRect(&rect);
	rect += rectClient.BottomRight() - rect.TopLeft() - m_posClose;
	m_wndClose.MoveWindow(rect);
	m_wndClose.Invalidate();

	// Move 'Refresh' button
	m_wndRefresh.GetWindowRect(&rect);
	rect += rectClient.BottomRight() - rect.TopLeft() - m_posRefresh;
	m_wndRefresh.MoveWindow(rect);
	m_wndRefresh.Invalidate();

	// Move checkbox
	m_wndUseOle.GetWindowRect(&rect);
	rect = CRect(CPoint(m_posUseOle.cx, rectClient.bottom - m_posUseOle.cy), rect.Size());
	m_wndUseOle.MoveWindow(rect);
	m_wndUseOle.Invalidate();

	UpdateWindow();
}

void CClipSpyDlg::SetColumns(BOOL fShowTymed) 
{
	// Delete any existing columns
	while (m_wndList.DeleteColumn(0))
		;

	// Set up report columns
	int nCol = 0;
	m_wndList.InsertColumn(nCol++, TEXT("Name"), LVCFMT_LEFT, LVSCW_AUTOSIZE_USEHEADER);
	m_wndList.InsertColumn(nCol++, TEXT("Format"), LVCFMT_LEFT, LVSCW_AUTOSIZE_USEHEADER);
	if (fShowTymed)
		m_wndList.InsertColumn(nCol++, TEXT("Medium"), LVCFMT_LEFT, LVSCW_AUTOSIZE_USEHEADER);
	m_wndList.InsertColumn(nCol++, TEXT("Size"), LVCFMT_RIGHT, LVSCW_AUTOSIZE_USEHEADER);
}

void CClipSpyDlg::Refresh() 
{
	UINT nFormat = 0;
	CString strNum;

	BOOL fUseOleClipboard = m_fUseOleClipboard || m_pDataObject;

	// Clear lists
	m_wndList.DeleteAllItems();
	m_formats.RemoveAll();
	SetColumns(fUseOleClipboard);

	// Open the clipboard!
	COleDataObject dataObject;
	COleDataObject* pDataObject = m_pDataObject;

	if (fUseOleClipboard)
	{
		if (NULL == pDataObject)
		{
			pDataObject = &dataObject;
			if (!pDataObject->AttachClipboard())
			{
				m_wndList.InsertItem(0, TEXT("Clipboard is busy"));
				return;
			}
		}
	}
	else
	{
		if (!OpenClipboard())
			return;
	}

	// Enumerate the formats on the clipboard
	FORMATETC fmt; ZeroMemory(&fmt, sizeof(fmt));
	if (fUseOleClipboard)
		pDataObject->BeginEnumFormats();

	for (int nIndex = 0;; nIndex++) {
		int nCurCol = 1; // Current column

		if (fUseOleClipboard)
		{
			if (!pDataObject->GetNextFormat(&fmt)) break;
			nFormat = fmt.cfFormat;
		}
		else
		{
			nFormat = EnumClipboardFormats(nFormat);
			if (0 == nFormat)
				break;
		}

		// Create a description of the format
		m_wndList.InsertItem(nIndex, GetFormatDescription(nFormat));
		strNum.Format(TEXT("0x%04x"), nFormat);
		m_wndList.SetItemText(nIndex, nCurCol++, strNum);

		if (fUseOleClipboard)
			m_wndList.SetItemText(nIndex, nCurCol++, GetMediumDescription(fmt.tymed));

		// Show the size if the item is an HGLOBAL
		HGLOBAL hData = NULL;
		if (fUseOleClipboard)
		{
			hData = pDataObject->GetGlobalData(nFormat);
		}
		else
		{
			if (IsMemoryFormat(nFormat))
				hData = GetClipboardData(nFormat);
		}
		if (NULL != hData)
			strNum.Format(TEXT("%d"), GlobalSize(hData));
		else
			strNum = TEXT("N/A");
		
		m_wndList.SetItemText(nIndex, nCurCol, strNum);

		m_formats.Add(nFormat);
	}

	if (!fUseOleClipboard)
		CloseClipboard();

	// Autosize columns
	LV_COLUMN lvc;
	lvc.mask = LVCF_TEXT | LVCF_WIDTH;
	TCHAR szText[256];
	lvc.pszText = szText;
	lvc.cchTextMax = sizeof(szText)/sizeof(szText[0]);
	for (int i = 0; i < 4; i++)
	{
		const int nMagicUndocumentedPadding = 12; // Bloody listview
		m_wndList.SetColumnWidth(i, LVSCW_AUTOSIZE);
		m_wndList.GetColumn(i, &lvc);
		int nWidth = m_wndList.GetStringWidth(lvc.pszText) + nMagicUndocumentedPadding;
		if (nWidth > lvc.cx)
			m_wndList.SetColumnWidth(i, nWidth);
	}
}

void CClipSpyDlg::OnClose() 
{
	CDialog::OnClose();
}

void CClipSpyDlg::OnDestroy() 
{
	CDialog::OnDestroy();

	// Remove from clipboard chain
	ChangeClipboardChain(m_hwndClipChainNext);

	// Save window position
	SavePersistentData();

	// Delete temp files
	while (!m_listTempFiles.IsEmpty())
		try { CFile::Remove(m_listTempFiles.RemoveHead()); }
		catch(CFileException* e) { e->Delete(); }
}

void CClipSpyDlg::OnDrawClipboard() 
{
	CDialog::OnDrawClipboard();
	// Pass message to next in chain
	if (NULL != m_hwndClipChainNext)
		::SendMessage(m_hwndClipChainNext, WM_DRAWCLIPBOARD, 0, 0);

	// Refresh!
	Refresh();	
}

void CClipSpyDlg::OnChangeCbChain(HWND hWndRemove, HWND hWndAfter) 
{
	CDialog::OnChangeCbChain(hWndRemove, hWndAfter);
	
	// Update 'next in chain' window handle
	m_hwndClipChainNext = hWndAfter;
}

void CClipSpyDlg::OnGetMinMaxInfo(MINMAXINFO FAR* lpMMI) 
{
	// Restrict tracking width
	if (m_fDoneInitDialog)
		lpMMI->ptMinTrackSize = m_ptMinSize;
	
	CDialog::OnGetMinMaxInfo(lpMMI);
}

BOOL CClipSpyDlg::SaveClipboardData(UINT nFormat, const CString& strFile)
{
	BOOL fSuccess = FALSE;

	HGLOBAL hData = NULL;
	COleDataObject dataObject;
	if (m_fUseOleClipboard)
	{
		if (!dataObject.AttachClipboard())
			return FALSE;
		hData = dataObject.GetGlobalData(nFormat);
	}
	else
	{
		if (!OpenClipboard())
			return FALSE;
		if (IsMemoryFormat(nFormat))
			hData = GetClipboardData(nFormat);
	}

	if (hData)
	{
		BYTE* pData = (BYTE*)GlobalLock(hData);
		try
		{
			CFile file(strFile, CFile::modeCreate|CFile::modeWrite);
			file.Write(pData, GlobalSize(hData));
			fSuccess = TRUE;
		}
		catch (CFileException* e)
		{
			e->ReportError();
			e->Delete();
		}
		GlobalUnlock(hData);
	}

	if (!m_fUseOleClipboard)
		CloseClipboard();

	return fSuccess;
}

BOOL CClipSpyDlg::SaveClipboardDataToTempFile(UINT nFormat, CString& strTempFile)
{
	CString strTempPath;
	GetTempPath(MAX_PATH, strTempPath.GetBuffer(MAX_PATH));
	strTempPath.ReleaseBuffer();
	GetTempFileName(strTempPath, "tmp", 0, strTempFile.GetBuffer(MAX_PATH));
	strTempFile.ReleaseBuffer();
	if (SaveClipboardData(nFormat, strTempFile))
	{
		// Record the tempfile name
		m_listTempFiles.AddHead(strTempFile);
		return TRUE;
	}
	return FALSE;
}

static HGLOBAL CreateDropSourceData(CString sFile)
{
	HGLOBAL hGlobal;
	
	// allocate space for DROPFILE structure plus the number of 
	// file and one extra byte for final NULL terminator
	hGlobal = GlobalAlloc(GHND|GMEM_SHARE,(DWORD) 
	(sizeof(DROPFILES)+(_MAX_PATH)+1));
	
	if(hGlobal == NULL)
		return hGlobal;

	LPDROPFILES pDropFiles;
	pDropFiles = (LPDROPFILES)GlobalLock(hGlobal);
	
	// set the offset where the starting point of the file start
   	pDropFiles->pFiles = sizeof(DROPFILES);
	
	// file contains wide characters
   	pDropFiles->fWide=FALSE;

	int CurPosition = sizeof(DROPFILES);
	lstrcpy((LPSTR)((LPSTR)(pDropFiles)+CurPosition),TEXT(sFile));
	CurPosition += sFile.GetLength() + 1;

	// final null terminator as per CF_HDROP Format specs.
	((LPSTR)pDropFiles)[CurPosition]=0;
	GlobalUnlock(hGlobal);
 	
	return hGlobal;
}

static HGLOBAL Copy(HGLOBAL hData)
{
	// Yawn
	DWORD dwSize = GlobalSize(hData);
	HGLOBAL hCopy = GlobalAlloc(GHND, dwSize);
	LPBYTE pData = (LPBYTE)GlobalLock(hData);
	LPBYTE pCopy = (LPBYTE)GlobalLock(hCopy);
	CopyMemory(pCopy, pData, dwSize);
	GlobalUnlock(hData);
	GlobalUnlock(hCopy);
	return hCopy;
}

void CClipSpyDlg::OnBegindragList(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_LISTVIEW* pNMListView = (NM_LISTVIEW*)pNMHDR;

	const UINT nFormat = m_formats[pNMListView->iItem];
	COleDataSource dataSource;
	BOOL fDoDragDrop = FALSE;

	// Save the clipboard format as a temporary file and do a HDROP
	CString strTempFile;
	if (SaveClipboardDataToTempFile(nFormat, strTempFile))
	{
		HGLOBAL hGlobal = CreateDropSourceData(strTempFile);
		dataSource.CacheGlobalData(CF_HDROP, hGlobal);
		fDoDragDrop = TRUE;
	}

	// We can offer it as its actual format too, but that is bad in certain cases,
	// e.g. for CF_METAFILEPICT, the METAFILEPICT contains a metafile handle 
	// which needs to be copied. In general we don't know what needs copying.
	if (m_fUseOleClipboard)
	{
		COleDataObject dataObject;
		if (dataObject.AttachClipboard())
		{
			switch (nFormat)
			{
			case CF_METAFILEPICT:
				{
					if (HGLOBAL hData = dataObject.GetGlobalData(nFormat))
					{
						METAFILEPICT* pMFP = (METAFILEPICT*)GlobalLock(hData);
						if (HMETAFILE hmfCopy = CopyMetaFile(pMFP->hMF, NULL))
						{
							if (HGLOBAL hCopy = Copy(hData))
							{
								pMFP = (METAFILEPICT*)GlobalLock(hCopy);
								pMFP->hMF = hmfCopy;
								GlobalUnlock(hCopy);
								STGMEDIUM stgm; ZeroMemory(&stgm, sizeof(stgm));
								stgm.tymed = TYMED_MFPICT;
								stgm.hMetaFilePict = hCopy;
								dataSource.CacheData(nFormat, &stgm);
								fDoDragDrop = TRUE;
							}
							else
								DeleteMetaFile(hmfCopy);
						}
						GlobalUnlock(hData);
					}
					break;
				}
			case CF_DIB:
				{
					if (HGLOBAL hData = dataObject.GetGlobalData(nFormat))
					{
						if (HGLOBAL hCopy = Copy(hData))
						{
							dataSource.CacheGlobalData(nFormat, hCopy);
							fDoDragDrop = TRUE;
						}
					}
				}
				break;
			}
		}
	}

	// Go for it
	if (fDoDragDrop)
	{
		m_dropTarget.Revoke();
		DROPEFFECT de = dataSource.DoDragDrop(DROPEFFECT_COPY);
		m_dropTarget.Register(this);
	}
}

void CClipSpyDlg::OnContextMenu(CWnd* pWnd, CPoint point) 
{
	// Figure out which item to use
	CPoint ptInList(point);
	m_wndList.ScreenToClient(&ptInList);
	int nSel;
	if (CPoint(-1, -1) == point) // Menu key?
	{
		// Menu key pressed - use the current selection
		nSel = m_wndList.GetNextItem(-1, LVNI_SELECTED);
		if (nSel < 0)
			return;

		// Get the centre of the item rect
		CRect rect;
		m_wndList.GetItemRect(nSel, rect, LVIR_LABEL);
		ptInList = rect.CenterPoint();
	}
	else
	{
		// Right button click - hit test the listview
		nSel = m_wndList.HitTest(ptInList);
		if (nSel < 0)
			return;
	}
	// Back to screen coords
	m_wndList.ClientToScreen(&ptInList);
	
	// We can only save HGLOBAL media
	BOOL fEnable = FALSE;

	// Check medium is HGLOBAL
	{
		COleDataObject dataObject;
		if (dataObject.AttachClipboard() &&
		  NULL != dataObject.GetGlobalData(m_formats[nSel]))
			fEnable = TRUE;
	}

	// Create context menu
	CCommandMenu menu;
	menu.Register(this);
	menu.AddToolbar(IDR_MAINFRAME);
	menu.CreatePopupMenu(NULL);
	menu.InsertMenu(-1, MF_BYPOSITION | (fEnable ? 0 : MF_GRAYED), 
		ID_FILE_SAVE_AS, TEXT("Save &As..."));
	menu.InsertMenu(-1, MF_BYPOSITION | MF_SEPARATOR);
	CFolderMenu* pmenuSendTo = new CFolderMenu;
	pmenuSendTo->CreatePopupMenu(&menu, TRUE);
	pmenuSendTo->AddFolderItems(ID_SENDTOFIRST, ID_SENDTOLAST, CSIDL_SENDTO);
	menu.InsertMenu(-1, MF_BYPOSITION | MF_POPUP | (fEnable ? 0 : MF_GRAYED), 
		(UINT)(HMENU)*pmenuSendTo, TEXT("Se&nd To"));

	// Do the menu thang
	UINT nID = menu.TrackPopupMenu(TPM_RIGHTBUTTON | TPM_RETURNCMD, 
		ptInList.x, ptInList.y, this);

	if (0 == nID)
	{
		return; // cancelled
	}
	else if (ID_FILE_SAVE_AS == nID)
	{
		CFileDialogX dlg(TRUE);
		dlg.SetProfileSection(m_strProfileSection);
		if (IDOK == dlg.DoModal())
			SaveClipboardData(m_formats[nSel], dlg.GetPathName());
	}
	else if (nID >= ID_SENDTOFIRST && nID < ID_SENDTOLAST)
	{
		BOOL fIsFolder = FALSE;
		CString strSendTo;
		if (pmenuSendTo->GetItemPath(nID, strSendTo, fIsFolder))
		{
			if (fIsFolder)
			{
				CFileDialog dlg(FALSE);
				dlg.m_ofn.lpstrInitialDir = strSendTo;
				if (IDOK == dlg.DoModal())
					SaveClipboardData(m_formats[nSel], dlg.GetPathName());
			}
			else
			{
				// Save the clipboard format as a temporary file
				CString strTempFile;
				if (!SaveClipboardDataToTempFile(m_formats[nSel], strTempFile))
					return;

				// Launch the temp file using the SendTo item
				CString strCmd; 
				strCmd.Format(TEXT("\"%s\" \"%s\""), strSendTo, strTempFile);
				if (32 > WinExec(strCmd, SW_SHOW))
//					if (32 > (int)ShellExecute(*this, NULL, strSendTo, strTempFile, NULL, SW_SHOW))
					AfxMessageBox(strSendTo + CString("\n\nCannot launch this file."));
			}
		}
	}
	else ASSERT(FALSE);
}

void CClipSpyDlg::OnInitMenuPopup(CMenu* pPopupMenu, UINT nIndex, BOOL bSysMenu) 
{
	if (bSysMenu)
	{
		pPopupMenu->CheckMenuItem(IDM_ALWAYSONTOP, 
			m_fAlwaysOnTop ? MF_CHECKED : MF_UNCHECKED);
	}

	CDialog::OnInitMenuPopup(pPopupMenu, nIndex, bSysMenu);
}

void CClipSpyDlg::OnButton1() 
{
	OleSetClipboard(NULL);
}

void CClipSpyDlg::OnCheckOleClipboard()
{
	UpdateData();
	Refresh();
}
