// FolderMenu.cpp
//
// Copyright (c) 1997, Jim Barry

#include "stdafx.h"
#include "FolderMenu.h"

// First some helpers
namespace
{
	// Get a CString out of a STRRET
	CString ExtractString(STRRET& name, LPCITEMIDLIST pidl)
	{
		USES_CONVERSION; // for ATL string conversion macros
		LPCTSTR lpName = NULL;
		switch (name.uType)
		{
		// Extract the string from the STRRET
		case STRRET_WSTR: 
			lpName = OLE2CT(name.pOleStr); 
			break;

		case STRRET_OFFSET: 
			ASSERT(NULL != pidl);
			lpName = A2CT(&((char*)&pidl->mkid)[name.uOffset]); // yuk
			break;

		case STRRET_CSTR: 
			lpName = A2CT(name.cStr); 
			break;

		default: 
			ASSERT(FALSE);
		}
		return lpName;
	}

	// Get the small icon from a pidl
	HICON ShiteExtractSmallIcon(IShellFolder* pFolder, LPCITEMIDLIST pidl)
	{
		SHFILEINFO shfi; ZeroMemory(&shfi, sizeof(shfi));
		SHGetFileInfo((LPCTSTR)pidl, 0, &shfi, sizeof(shfi), 
			SHGFI_ICON | SHGFI_SMALLICON | SHGFI_PIDL);
		return shfi.hIcon;
	}

	HICON ExtractSmallIcon(IShellFolder* pFolder, LPCITEMIDLIST pidl)
	{
		HICON hicon = NULL;
		CComPtr<IExtractIcon> pExtractIcon;
		if (SUCCEEDED(pFolder->GetUIObjectOf(*AfxGetMainWnd(), 1,
			(LPCITEMIDLIST*)&pidl, IID_IExtractIcon, NULL, (LPVOID*)&pExtractIcon)))
		{
			int nIndex = 0;
			UINT wFlags = 0;
			char szIconFile[MAX_PATH];
			if (S_OK == pExtractIcon->GetIconLocation(GIL_FORSHELL, 
				szIconFile, sizeof(szIconFile), &nIndex, &wFlags))
			{
				HICON hiconLarge = NULL;
				HICON hiconSmall = NULL;
				if (S_OK != (pExtractIcon->Extract(szIconFile, nIndex, 
					&hiconLarge, &hiconSmall, MAKELONG(32, 16))))
				{
					// Try it this way instead
					UINT nResult = ExtractIconEx(szIconFile, nIndex, 
						NULL, &hiconSmall, 1);

					if (NULL == hiconSmall)
						TRACE("ExtractSmallIcon: failed\n");
				}
				if (hiconSmall)
					hicon = hiconSmall;
				if (hiconLarge)
					DestroyIcon(hiconLarge);
			}
		}

		return hicon;
	}

	// Resolve a shell link
	BOOL ResolveLink(HWND hwnd, LPCSTR lpszLinkFile, CString& strPath, 
					 WIN32_FIND_DATA* pWFD = NULL)
	{
		USES_CONVERSION;
		HRESULT hr;
		WIN32_FIND_DATA wfd;
		if (NULL == pWFD)
			pWFD = &wfd;

		// Get a pointer to the IShellLink interface. 
		CComPtr<IShellLink> pShellLink; 
		CoCreateInstance(CLSID_ShellLink, NULL, 
			CLSCTX_INPROC_SERVER, IID_IShellLink, (LPVOID*)&pShellLink); 

		if (!pShellLink)
			return FALSE;

		CComQIPtr<IPersistFile, &IID_IPersistFile> pPersistFile(pShellLink);
		if (!pPersistFile)
			return FALSE;

		if (FAILED(pPersistFile->Load(T2OLE(lpszLinkFile), STGM_READ)))
			return FALSE;

		// Resolve the link. 
		if (FAILED(pShellLink->Resolve(hwnd, SLR_ANY_MATCH)))
			return FALSE;

		// Get the path to the link target.
		hr = pShellLink->GetPath(strPath.GetBuffer(MAX_PATH), MAX_PATH, 
			pWFD, SLGP_UNCPRIORITY); 
		strPath.ReleaseBuffer();

		return SUCCEEDED(hr);
	}
}

//
// CFolderMenu
//

CFolderMenu::CFolderMenu()
{
	m_idCmdFirst = 0;
	m_idCmdLast = 0;
	m_idCmdNext = 0;
	m_fDoneInit = FALSE;
}

CFolderMenu::CPidlArray& CFolderMenu::GetPidls()
{
	// Walk up menu hierarchy
	CFolderMenu* pOwner = dynamic_cast<CFolderMenu*>(m_pOwner);
	return pOwner ? pOwner->GetPidls() : m_rgPidls;
}

UINT& CFolderMenu::GetCmdFirst()
{
	// Walk up menu hierarchy
	CFolderMenu* pOwner = dynamic_cast<CFolderMenu*>(m_pOwner);
	return pOwner ? pOwner->GetCmdFirst() : m_idCmdFirst;
}

UINT& CFolderMenu::GetCmdLast()
{
	// Walk up menu hierarchy
	CFolderMenu* pOwner = dynamic_cast<CFolderMenu*>(m_pOwner);
	return pOwner ? pOwner->GetCmdLast() : m_idCmdLast;
}

UINT& CFolderMenu::GetCmdNext()
{
	// Walk up menu hierarchy
	CFolderMenu* pOwner = dynamic_cast<CFolderMenu*>(m_pOwner);
	return pOwner ? pOwner->GetCmdNext() : m_idCmdNext;
}

void CFolderMenu::SetCmdNext(UINT idCmd)
{
	// Walk up menu hierarchy
	CFolderMenu* pOwner = dynamic_cast<CFolderMenu*>(m_pOwner);
	if (pOwner)
		pOwner->SetCmdNext(idCmd);
	else
		m_idCmdNext = idCmd;
}

BOOL CFolderMenu::GetMalloc()
{
	// Lazy
	if (!m_pMalloc)
	{
		if (FAILED(SHGetMalloc(&m_pMalloc)))
			return FALSE;
	}
	return TRUE;
}

IShellFolder* CFolderMenu::GetRootFolder()
{
	// Walk up menu hierarchy
	CFolderMenu* pOwner = dynamic_cast<CFolderMenu*>(m_pOwner);
	return pOwner ? pOwner->GetRootFolder() : m_pRootFolder;
}

BOOL CFolderMenu::AddFolderItems(UINT idCmdFirst, UINT idCmdLast, int nFolder)
{
	// Get the task allocator (if we didn't already)
	if (!GetMalloc())
		return FALSE;

	// Get the pidl of the required special folder
	LPITEMIDLIST pidl;
	if (NOERROR != SHGetSpecialFolderLocation(*AfxGetMainWnd(), nFolder, &pidl))
		return FALSE;

	// Call the other overload of AddFolderItems
	BOOL fResult = AddFolderItems(idCmdFirst, idCmdLast, pidl);
	m_pMalloc->Free(pidl);
	return fResult;
}

BOOL CFolderMenu::AddFolderItems(UINT idCmdFirst, UINT idCmdLast, LPCITEMIDLIST pidl)
{
	m_idCmdFirst = idCmdFirst;
	m_idCmdLast = idCmdLast;
	m_idCmdNext = idCmdFirst;

	// Get the task allocator (if we didn't already)
	if (!GetMalloc())
		return FALSE;

	// Get the desktop's IShellFolder
	CComPtr<IShellFolder> pDesktopFolder;
	if (FAILED(SHGetDesktopFolder(&pDesktopFolder)))
		return FALSE;

	// Get the target IShellFolder
	if (FAILED(pDesktopFolder->BindToObject(
	  pidl, NULL, IID_IShellFolder, (LPVOID*)&m_pRootFolder)))
		return FALSE;

	//	Folder items will be added in OnInitPopupMenu

	return TRUE;
}

int CFolderMenu::AddFolderItems()
{
	USES_CONVERSION;

	// Did we do this already?
	if (m_fDoneInit)
		return 0;
	m_fDoneInit = TRUE;

	// Get the task allocator (if we didn't already)
	if (!GetMalloc())
		return 0;

	// Enumerate the folder items 
	CStringArray rgNames[2]; // folder, non-folder
	CPidlArray rgPidls[2]; // folder, non-folder
	STRRET name;
	LPITEMIDLIST pidl = NULL;
	const int iPidlBase = GetPidls().GetSize();
	int nNonFolders = 0;

	// Get the folder enumerator
	CComPtr<IEnumIDList> pEnum;
	m_pRootFolder->EnumObjects(NULL, SHCONTF_FOLDERS|SHCONTF_NONFOLDERS, &pEnum);
	if (!pEnum)
		return 0;

	// Enumerate folder items
	while (S_OK == pEnum->Next(1, &pidl, NULL))
	{
		name.uType = STRRET_WSTR;
		if (SUCCEEDED(m_pRootFolder->GetDisplayNameOf(pidl, SHGDN_NORMAL, &name)))
		{
			CString strName(ExtractString(name, pidl));

			// Is this a folder or what
			ULONG flags = SFGAO_CONTENTSMASK;
			m_pRootFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidl, &flags);
			int iType = ((SFGAO_HASSUBFOLDER|SFGAO_FOLDER) & flags) ? 0 : 1;

			// Alphabetize the items
			int nSize = rgNames[iType].GetSize();
			for (int i = 0; i < nSize; i++)
				if (strName.CompareNoCase(rgNames[iType][i]) < 0)
					break;

			// Store the name and pidl
			rgNames[iType].InsertAt(i, strName);
			rgPidls[iType].InsertAt(i, pidl);
		}
	}

	// Pidl array will correspond to item IDs
	int iPidl = GetCmdNext() - GetCmdFirst();
	int iMenuIndex = 0;

	// Add the folders to the SendTo menu
	int nSize = rgNames[0].GetSize();
	for (int i = 0; i < nSize; i++)
	{
		if (iPidl + GetCmdFirst() >= GetCmdLast())
			break;

		// Get the subfolder object
		CComPtr<IShellFolder> pSubFolder;
		if (SUCCEEDED(m_pRootFolder->BindToObject(
		  rgPidls[0][i], NULL, IID_IShellFolder, (LPVOID*)&pSubFolder)))
		{
			CFolderMenu* pSubMenu = new CFolderMenu;
			pSubMenu->CreatePopupMenu(this, TRUE);
			pSubMenu->m_pRootFolder = pSubFolder;
			InsertMenu(iMenuIndex, MF_POPUP|MF_BYPOSITION, (UINT)(HMENU)*pSubMenu, rgNames[0][i]);
			SetIcon(iMenuIndex++, ExtractSmallIcon(m_pRootFolder, rgPidls[0][i]));
		}
		else
			TRACE("Cannot bind to pidl %x \"%s\"\n", rgPidls[0][i], rgNames[0][i]);
	}

	// Now the non-folders
	nSize = rgNames[1].GetSize();
	for (i = 0; i < nSize; i++)
	{
		if (iPidl + GetCmdFirst() > GetCmdLast())
			break;

		// Insert the item
		InsertMenu(iMenuIndex, MF_BYPOSITION, iPidl + GetCmdFirst(), rgNames[1][i]);
		SetIcon(iMenuIndex++, ExtractSmallIcon(m_pRootFolder, rgPidls[1][i]));

		// Translate relative pidl to absolute
		LPITEMIDLIST pidl = rgPidls[1][i];
		if (SUCCEEDED(m_pRootFolder->GetDisplayNameOf(pidl, SHGDN_FORPARSING, &name)))
		{
			CString strParseName(ExtractString(name, pidl));
			ULONG chEaten, dwAttributes;
			pidl = NULL;
			CComPtr<IShellFolder> pDesktopFolder;
			SHGetDesktopFolder(&pDesktopFolder);
			if (FAILED(pDesktopFolder->ParseDisplayName(NULL, NULL, T2OLE(strParseName), 
				&chEaten, &pidl, &dwAttributes)))
			{
				TRACE("ParseDisplayName: failed for %s\n", strParseName);
			}
		}
		else
			TRACE("GetDisplayNameOf: failed\n");

		// Stick it in the array
		GetPidls().InsertAt(iPidl++, pidl);
	}

	// Create a disabled 'empty' item if the folder has no items
	if (0 == GetMenuItemCount())
	{
		// Hack: Insert a dummy pidl entry for the "(empty)" item so that
		// we have unique menu IDs (MFC owner-draw relies on this)
		GetPidls().InsertAt(iPidl, (LPITEMIDLIST)NULL);
		InsertMenu(0, MF_BYPOSITION|MF_GRAYED, iPidl + GetCmdFirst(), TEXT("(Empty)"));
		iPidl++; // counts as a non-folder for our purposes
	}

	// Store the next available command ID
	SetCmdNext(GetCmdFirst() + iPidl);

	return 0;
}

BOOL CFolderMenu::GetItemPath(UINT idCmd, CString& strPath, BOOL& fIsFolder, 
	BOOL fResolveLinks)
{
	LPITEMIDLIST pidl;
	if (!GetItemIDList(idCmd, &pidl))
		return FALSE;

	// Get the desktop's IShellFolder
	CComPtr<IShellFolder> pDesktopFolder;
	if (FAILED(SHGetDesktopFolder(&pDesktopFolder)))
		return FALSE;

	// Extract the folder item's path
	STRRET name;
	name.uType = STRRET_WSTR;
	if (FAILED(pDesktopFolder->GetDisplayNameOf(pidl, SHGDN_FORPARSING, &name)))
		return FALSE;
	strPath = ExtractString(name, pidl);

	// Get the folder item's attributes
	ULONG flags = SFGAO_DISPLAYATTRMASK | SFGAO_CONTENTSMASK;
	if (FAILED(pDesktopFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidl, &flags)))
		return FALSE;
	fIsFolder = (SFGAO_FOLDER & flags);

	// Follow links, if required
	if (fResolveLinks) // && (SFGAO_LINK & flags))
	{
		// Is it a shell link
		WIN32_FIND_DATA fd;
		CString strLinkTarget;
		if (ResolveLink(*AfxGetMainWnd(), strPath, strLinkTarget, &fd))
		{
			fIsFolder = (FILE_ATTRIBUTE_DIRECTORY & fd.dwFileAttributes);
			strPath = strLinkTarget;
		}
	}

	return TRUE;
}

BOOL CFolderMenu::GetItemIDList(UINT idCmd, LPITEMIDLIST* ppidl)
{
	// Check ID is in range
	if (idCmd < GetCmdFirst() || idCmd > GetCmdLast())
		return FALSE;

	// Go get it
	*ppidl = GetPidls()[idCmd - GetCmdFirst()];
	return NULL != *ppidl;
}

void CFolderMenu::OnInitMenuPopup(UINT nIndex, BOOL bSysMenu)
{
	// Add folder items 'on demand'
	AddFolderItems();
	ProcessMenuItems(TRUE); // does the column breaking
}

void CFolderMenu::OnMenuClose()
{
	CImageMenu::OnMenuClose();
}
