// FolderMenu.h
//
// Copyright (c) 1997, Jim Barry

#ifndef _FolderMenu_H_
#define _FolderMenu_H_

#include "ImageMenu.h"
#include <shlobj.h>

// CFolderMenu implements "Start Menu" style cascading menus.
// Create a submenu with CreatePopupMenu then call AddFolderItems, either supplying
// an ITEMIDLIST or a special folder value as listed for SHGetSpecialFolderLocation.
// After calling TrackPopupMenu (perhaps with the TMP_RETURNCMD flag), retrieve 
// the selected item's path using GetItemPath, or retrieve its ITEMIDLIST using 
// GetItemIDList. Note that the ITEMIDLIST will be deleted when the menu object
// goes out of scope.
class CFolderMenu : public CIconMenu
{
public:
	CFolderMenu();
	BOOL AddFolderItems(UINT idCmdFirst, UINT idCmdLast, LPCITEMIDLIST pidl);
	BOOL AddFolderItems(UINT idCmdFirst, UINT idCmdLast, int nFolder);
	BOOL GetItemPath(UINT idCmd, CString& strPath, BOOL& fIsFolder, 
		BOOL fResolveLinks = TRUE);
	BOOL GetItemIDList(UINT idCmd, LPITEMIDLIST* ppidl);

// Overrides
protected:
	virtual void OnInitMenuPopup(UINT nIndex, BOOL bSysMenu);
	virtual void OnMenuClose();

// Implementation
protected:
	class CPidlArray : public CArray<LPITEMIDLIST, LPITEMIDLIST>
	{
	public:
		CPidlArray::~CPidlArray()
		{
			// Free the pidls!
			CComPtr<IMalloc> pMalloc;
			if (SUCCEEDED(SHGetMalloc(&pMalloc)))
				for (int i = 0; i < GetSize(); i++)
					pMalloc->Free(GetAt(i));
		}
	};

	CFolderMenu* GetRootMenu();
	int AddFolderItems();
	CPidlArray& GetPidls();
	UINT& GetCmdFirst();
	UINT& GetCmdLast();
	UINT& GetCmdNext();
	void SetCmdNext(UINT idCmd);
	BOOL GetMalloc();
	IShellFolder* GetRootFolder();

	CComPtr<IShellFolder> m_pRootFolder;
	CComPtr<IMalloc> m_pMalloc;
	CPidlArray m_rgPidls;
	UINT m_idCmdFirst;
	UINT m_idCmdLast;
	UINT m_idCmdNext;
	BOOL m_fDoneInit;
};

#endif // _FolderMenu_H_