// Persist.cpp
//
// Copyright (c) 1998 Jim Barry

#include "stdafx.h"
#include "persist.h"

// Require MFC common control classes
#include <afxcmn.h>

/////////////////////////////////
// CPersistence class

CPersistence::CPersistence(LPCTSTR lpProfileSection)
{
	m_strProfileSection = lpProfileSection;
	m_fStoring = FALSE;
}

CPersistence::~CPersistence()
{
}

void CPersistence::SetProfileSection(LPCTSTR lpProfileSection)
{
	m_strProfileSection = lpProfileSection;
}

BOOL CPersistence::ExchangePersistentData(BOOL fSave)
{
	// Set direction flag and call virtual exchange handler
	m_fStoring = fSave;
	OnExchangePersistentData();
	return TRUE;
}

void CPersistence::OnExchangePersistentData()
{
	// No Default
}

void CPersistence::PDX_String(LPCTSTR lpKey, CString& str, LPCTSTR lpDefault)
{
	if (IsStoring())
		// No error check yet
		AfxGetApp()->WriteProfileString(m_strProfileSection, lpKey, str);
	else
		str = AfxGetApp()->GetProfileString(m_strProfileSection, lpKey, lpDefault);
}

void CPersistence::PDX_Int(LPCTSTR lpKey, int& i, int nDefault)
{
	CString str;
	if (IsStoring())
	{
		str.Format(TEXT("%d"), i);
		PDX_String(lpKey, str);
	}
	else
	{
		CString strDefault;
		strDefault.Format(TEXT("%d"), nDefault);
		PDX_String(lpKey, str, strDefault);
		if (1 != _stscanf(str, TEXT("%d"), &i))
			i = nDefault;
	}
}

void CPersistence::PDX_Uint(LPCTSTR lpKey, UINT& i, UINT nDefault)
{
	CString str;
	if (IsStoring())
	{
		str.Format(TEXT("%u"), i);
		PDX_String(lpKey, str);
	}
	else
	{
		CString strDefault;
		strDefault.Format(TEXT("%d"), nDefault);
		PDX_String(lpKey, str, strDefault);
		_stscanf(str, TEXT("%u"), &i);
	}
}

void CPersistence::PDX_Word(LPCTSTR lpKey, WORD& w, WORD wDefault)
{
	CString str;
	if (IsStoring())
	{
		str.Format(TEXT("%hu"), w);
		PDX_String(lpKey, str);
	}
	else
	{
		CString strDefault;
		strDefault.Format(TEXT("%d"), wDefault);
		PDX_String(lpKey, str, strDefault);
		_stscanf(str, TEXT("%hu"), &w);
	}
}

void CPersistence::PDX_Dword(LPCTSTR lpKey, DWORD& i, DWORD nDefault)
{
	CString str;
	if (IsStoring())
	{
		str.Format(TEXT("%lu"), i);
		PDX_String(lpKey, str);
	}
	else
	{
		CString strDefault;
		strDefault.Format(TEXT("%lu"), nDefault);
		PDX_String(lpKey, str, strDefault);
		_stscanf(str, TEXT("%lu"), &i);
	}
}

void CPersistence::PDX_Bool(LPCTSTR lpKey, BOOL& f, BOOL fDefault)
{
	CString str;
	if (IsStoring())
	{
		str = f ? TEXT("1") : TEXT("0");
		PDX_String(lpKey, str);
	}
	else
	{
		// A list of strings which can be used in the ini file
		// to represent a boolean
		const static struct {
			TCHAR*	pszString;
			BOOL	fResult;
		} table[] = {
			TEXT("yes"),	TRUE,
			TEXT("no"),	FALSE,
			TEXT("on"),	TRUE,
			TEXT("off"),	FALSE,
			TEXT("1"),	TRUE,
			TEXT("0"),	FALSE,
			NULL,	NULL	// Sentinel entry
		};
		
		// Get string entry and convert to lower case
		CString str;
		PDX_String(lpKey, str);
		str.MakeLower();
		
		// Go through the list
		for (int i = 0; table[i].pszString; i++)
			if (0 == str.CompareNoCase(table[i].pszString))
			{
				// Found a match
				f = table[i].fResult;
				return;
			}
		
		// return default if no match found
		f =  fDefault;
	}
}

void CPersistence::PDX_Rect(LPCTSTR lpKey, RECT& rect, LPCRECT pDefault)
{
	CString str;
	if (IsStoring())
	{
		str.Format(TEXT("%d,%d,%d,%d"), rect.left, rect.top, 
			rect.right, rect.bottom);
		PDX_String(lpKey, str);
	}
	else
	{
		rect = CRect(pDefault);
		PDX_String(lpKey, str);
		_stscanf(str, TEXT("%d,%d,%d,%d"), &rect.left, &rect.top, 
			&rect.right, &rect.bottom);
	}
}

void CPersistence::PDX_Point(LPCTSTR lpKey, POINT& pt, LPPOINT pDefault)
{
	CString str;
	if (IsStoring())
	{
		str.Format(TEXT("%d,%d"), pt.x, pt.y);
		PDX_String(lpKey, str);
	}
	else
	{
		pt = CPoint(*pDefault);
		PDX_String(lpKey, str);
		_stscanf(str, TEXT("%d,%d"), &pt.x, &pt.y);
	}
}

void CPersistence::PDX_Size(LPCTSTR lpKey, SIZE& size, LPSIZE lpDefault)
{
	CString str;
	if (IsStoring())
	{
		str.Format(TEXT("%d,%d"), size.cx, size.cy);
		PDX_String(lpKey, str);
	}
	else
	{
		size = CSize(*lpDefault);
		PDX_String(lpKey, str);
		_stscanf(str, TEXT("%d,%d"), &size.cx, &size.cy);
	}
}

void CPersistence::PDX_WindowPos(LPCTSTR lpKey, CWnd* pWnd, LPPOINT lpDefault)
{
	if (!pWnd || !pWnd->GetSafeHwnd())
	{
		ASSERT(FALSE);
		return;
	}

	CString str;
	CPoint pt;
	CRect rect;
	if (IsStoring())
	{
		pWnd->GetWindowRect(rect);
		PDX_Point(lpKey, rect.TopLeft());
	}
	else
	{
		PDX_String(lpKey, str);
		if (2 != _stscanf(str, TEXT("%d,%d"), &pt.x, &pt.y))
		{
			// Revert to default
			if (lpDefault)
				pt = CPoint(*lpDefault);
			else
				// No value or default - bail out
				return;
		}
		
		// Make sure it's on the desktop
		CWnd::GetDesktopWindow()->GetClientRect(rect);
		if (pt.x > rect.right || pt.y > rect.bottom)
			// Off-screen, center it instead
			pWnd->CenterWindow();
		else
			// OK, move the window
			pWnd->SetWindowPos(NULL, pt.x, pt.y, 0, 0, SWP_NOSIZE);
	}
}

void CPersistence::PDX_WindowSize(LPCTSTR lpKey, CWnd* pWnd, LPSIZE lpDefault)
{
	if (!pWnd || !pWnd->GetSafeHwnd())
	{
		ASSERT(FALSE);
		return;
	}

	CString str;
	CSize size;
	CRect rect;
	if (IsStoring())
	{
		pWnd->GetWindowRect(rect);
		PDX_Size(lpKey, rect.Size());
	}
	else
	{
		PDX_String(lpKey, str);
		if (2 != _stscanf(str, TEXT("%d,%d"), &size.cx, &size.cy))
		{
			// Revert to default
			if (lpDefault)
				size = CSize(*lpDefault);
			else
				// No value or default - bail out
				return;
		}
		
		// OK, move the window
		pWnd->SetWindowPos(NULL, 0, 0, size.cx, size.cy, SWP_NOMOVE);
	}
}

void CPersistence::PDX_WindowRect(LPCTSTR lpKey, CWnd* pWnd, LPCRECT lpDefault)
{
	if (!pWnd || !pWnd->GetSafeHwnd())
	{
		ASSERT(FALSE);
		return;
	}

	CString str;
	CRect rect;
	
	BOOL fMax = FALSE;
	BOOL fMin = FALSE;
	WINDOWPLACEMENT wp;
	wp.length = sizeof(wp);
	pWnd->GetWindowPlacement(&wp);

	if (IsStoring())
	{
		fMax = (wp.flags & WPF_RESTORETOMAXIMIZED) ||
			(SW_SHOWMAXIMIZED == wp.showCmd);
		fMin = (SW_SHOWMINIMIZED == wp.showCmd);
		rect = wp.rcNormalPosition;
		str.Format(TEXT("%d,%d,%d,%d,%d,%d"), rect.left, rect.top, 
			rect.right, rect.bottom, fMax, fMin);
		PDX_String(lpKey, str);
	}
	else
	{
		PDX_String(lpKey, str);
		if (4 > _stscanf(str, TEXT("%d,%d,%d,%d,%d,%d"), &rect.left, &rect.top, 
		  &rect.right, &rect.bottom, &fMax, &fMin))
		{
			// Revert to default
			if (lpDefault)
				rect = CRect(*lpDefault);
			else
				// No value or default - bail out
				return;
		}

		// Set the new placement
		wp.length = sizeof(wp);
		wp.rcNormalPosition = rect;

		// Set up min/max states
		if (fMin)
		{
			wp.showCmd = SW_SHOWMINIMIZED;
			if (fMax)
				wp.flags |= WPF_RESTORETOMAXIMIZED;
		}
		else if (fMax)
			wp.showCmd = SW_SHOWMAXIMIZED;

		pWnd->SetWindowPlacement(&wp);
	}
}

void CPersistence::PDX_MainFrameRect(LPCTSTR lpKey, CFrameWnd* pWnd, LPCRECT lprcDefault)
{
	if (!pWnd || !pWnd->GetSafeHwnd())
	{
		ASSERT(FALSE);
		return;
	}

	CString str;
	CRect rect;
	
	BOOL fMax = FALSE;
	WINDOWPLACEMENT wp;
	wp.length = sizeof(wp);
	pWnd->GetWindowPlacement(&wp);

	if (IsStoring())
	{
		fMax = (wp.flags & WPF_RESTORETOMAXIMIZED) ||
			(SW_SHOWMAXIMIZED == wp.showCmd);
		rect = wp.rcNormalPosition;
		str.Format(TEXT("%d,%d,%d,%d,%d"), rect.left, rect.top, 
			rect.right, rect.bottom, fMax);
		PDX_String(lpKey, str);
	}
	else
	{
		PDX_String(lpKey, str);
		if (4 > _stscanf(str, TEXT("%d,%d,%d,%d,%d"), &rect.left, &rect.top, 
		  &rect.right, &rect.bottom, &fMax))
		{
			// Revert to default
			if (lprcDefault)
				rect = CRect(*lprcDefault);
			else
				// No value or default - bail out
				return;
		}

		// Set the new placement
		wp.length = sizeof(wp);
		wp.rcNormalPosition = rect;
		
		// 'Disgusting' hack to get timing right for main window
		ASSERT_KINDOF(CFrameWnd, pWnd);
		wp.showCmd = SW_HIDE;
		if (fMax)
			AfxGetApp()->m_nCmdShow = SW_SHOWMAXIMIZED;

		pWnd->SetWindowPlacement(&wp);
	}
}

void CPersistence::PDX_MainFrameRect(LPCTSTR lpKey, CFrameWnd* pWnd, 
	const SIZE sizeDefault)
{
	if (!pWnd || !pWnd->GetSafeHwnd())
	{
		ASSERT(FALSE);
		return;
	}

	CString str;
	CRect rect;
	
	BOOL fMax = FALSE;
	WINDOWPLACEMENT wp;
	wp.length = sizeof(wp);
	pWnd->GetWindowPlacement(&wp);

	if (IsStoring())
	{
		fMax = (wp.flags & WPF_RESTORETOMAXIMIZED) ||
			(SW_SHOWMAXIMIZED == wp.showCmd);
		rect = wp.rcNormalPosition;
		str.Format(TEXT("%d,%d,%d,%d,%d"), rect.left, rect.top, 
			rect.right, rect.bottom, fMax);
		PDX_String(lpKey, str);
	}
	else
	{
		PDX_String(lpKey, str);
		if (4 > _stscanf(str, TEXT("%d,%d,%d,%d,%d"), &rect.left, &rect.top, 
		  &rect.right, &rect.bottom, &fMax))
		{
			// Revert to default size
			pWnd->SetWindowPos(NULL, 0, 0, sizeDefault.cx, sizeDefault.cy,
				SWP_NOMOVE | SWP_NOZORDER);
			return;
		}

		// Set the new placement
		wp.length = sizeof(wp);
		wp.rcNormalPosition = rect;
		
		// 'Disgusting' hack to get timing right for main window
		ASSERT_KINDOF(CFrameWnd, pWnd);
		wp.showCmd = SW_HIDE;
		if (fMax)
			AfxGetApp()->m_nCmdShow = SW_SHOWMAXIMIZED;

		pWnd->SetWindowPlacement(&wp);
	}
}

void CPersistence::PDX_BarState(LPCTSTR lpKey, CFrameWnd* pWnd)
{
	// This is way too easy!
	IsStoring() ? pWnd->SaveBarState(lpKey) : pWnd->LoadBarState(lpKey);
}

void CPersistence::PDX_IntVec(LPCTSTR lpKey, int* pVec, UINT nCount)
{
	ASSERT(pVec && (nCount > 0));

	CString str;
	if (IsStoring())
	{
		CString strTemp;
		for (UINT i = 0; i < nCount; i++)
		{
			LPCTSTR pszFormat = (i == (nCount - 1)) ? TEXT("%d") : TEXT("%d,");
			strTemp.Format(pszFormat, pVec[i]);
			str += strTemp;
		}
		PDX_String(lpKey, str);
	}
	else
	{
		PDX_String(lpKey, str);
		LPCTSTR pszBuffer = str;
		for (UINT i = 0; i < nCount; i++)
		{
			int nChars = 0;
			LPCTSTR pszFormat = (i == (nCount - 1)) ? TEXT("%d%n") : TEXT("%d,%n");
			if (1 > _stscanf(pszBuffer, pszFormat, &pVec[i], &nChars))
				return;
			pszBuffer += nChars;
		}
	}
}

// Exchange binary data, using hexadecimal string format.
// When reading, if cbData does not match the amount of data available, 
// then no data is copied.
void CPersistence::PDX_Data(LPCTSTR lpKey, void* pData, UINT cbData)
{
	CString strData, strTemp;
	const UINT cchTag = sizeof(UINT) * 2;
	UINT i;

	if (IsStoring())
	{
		// Convert the length, using machine byte order
		for (i = 0; i < sizeof(cbData); i++)
		{
			strTemp.Format(TEXT("%02x"), ((BYTE*)&cbData)[i]);
			strData += strTemp;
		}
		// Convert the data
		for (UINT i = 0; i < cbData; i++)
		{
			strTemp.Format(TEXT("%02x"), ((BYTE*)pData)[i]);
			strData += strTemp;
		}
		// Save the key
		PDX_String(lpKey, strData);
	}
	else
	{
		// Read the data string
		PDX_String(lpKey, strData);

		// Get the data length, using machine byte order
		if (strData.GetLength() < cchTag)
			return; // eeek
		UINT cbDataHere = 0;
		for (i = 0; i < cchTag / 2; i++)
		{
			int byte;
			if (1 != _stscanf(&((LPCTSTR)strData)[i*2], TEXT("%2x"), &byte))
				return;
			reinterpret_cast<BYTE*>(&cbDataHere)[i] = byte;
		}
		if (cbDataHere != cbData)
			return; // wrong length

		// Read the data
		UINT nVal = 0;
		for (i = 0; i < cbData; i++)
		{
			if (1 != _stscanf(&((LPCTSTR)strData)[cchTag + i*2], TEXT("%2x"), &nVal))
				return;
			reinterpret_cast<BYTE*>(pData)[i] = nVal;
		}
	}
}

////////////////////////////
// CFileDialogX

IMPLEMENT_DYNAMIC(CFileDialogX, CFileDialog)

CFileDialogX::CFileDialogX(
	BOOL fOpen, LPCTSTR lpszDefExt, LPCTSTR lpszFileName,
	DWORD dwFlags, LPCTSTR lpszFilter, CWnd* pParentWnd)
	: CFileDialog(fOpen, lpszDefExt, lpszFileName, dwFlags,
		lpszFilter, pParentWnd),
	  CPersistence(NULL)
{
	// No help for now
	m_ofn.Flags &= ~OFN_SHOWHELP;
}

void CFileDialogX::SetDefaultInitialFolder(LPCTSTR lpszFolder)
{
	m_strInitialFolder = lpszFolder;
}

int CFileDialogX::DoModal()
{
	// Set the initial directory and file type index
	LoadPersistentData();
	m_ofn.lpstrInitialDir = m_strInitialFolder;
	m_ofn.nFilterIndex = m_nFilterIndex;
	
	// Launch dialog
	int nResult = CFileDialog::DoModal();
	
	// Save directory name
	if (IDOK == nResult)
	{
		if (POSITION pos = GetStartPosition())
		{
			m_strInitialFolder = GetNextPathName(pos);
			int nSlash = m_strInitialFolder.ReverseFind('\\');
			if (nSlash >= 0)
				m_strInitialFolder = m_strInitialFolder.Left(nSlash);
		}
		m_nFilterIndex = m_ofn.nFilterIndex;
		SavePersistentData();
	}
	
	return nResult;
}

void CFileDialogX::OnExchangePersistentData()
{
	PDX_String(TEXT("InitialFolder"), m_strInitialFolder, m_strInitialFolder);
	PDX_Int(TEXT("FilterIndex"), m_nFilterIndex);
}

/////////////////////////////////////////////////////////////////////////////
// CPropertySheetX

IMPLEMENT_DYNAMIC(CPropertySheetX, CPropertySheet)

CPropertySheetX::CPropertySheetX(UINT nIDCaption, CWnd* pParentWnd, UINT iSelectPage)
	:CPropertySheet(nIDCaption, pParentWnd, iSelectPage),
	 CPersistence(m_strCaption)
{
	// Set profile section to be same as property sheet caption
	//SetProfileSection(m_strCaption);
}

CPropertySheetX::CPropertySheetX(LPCTSTR pszCaption, CWnd* pParentWnd, UINT iSelectPage)
	:CPropertySheet(pszCaption, pParentWnd, iSelectPage),
	 CPersistence(m_strCaption)
{
	// Set profile section to be same as property sheet caption
	//SetProfileSection(m_strCaption);
}

void CPropertySheetX::OnExchangePersistentData()
{
	PDX_Int(TEXT("Last Page"), m_nPage, -1);
}

/////////////////////////////////////////////////////////////////////////////
// CPropertySheetX message handlers

BOOL CPropertySheetX::OnInitDialog() 
{
	CPropertySheet::OnInitDialog();

	LoadPersistentData();
	if (m_nPage >= 0)
		SetActivePage(m_nPage); // Set the initial page

	return TRUE;
}

BOOL CPropertySheetX::DestroyWindow() 
{
	// Save the last page number
	SavePersistentData();
	return CPropertySheet::DestroyWindow();
}

BOOL CPropertySheetX::OnNotify(WPARAM wParam, LPARAM lParam, LRESULT* pResult) 
{
	// Watch out for TCN_SELCHANGE notifications from the tab control
	NMHDR* pNMHDR = (NMHDR*)lParam;
	if (pNMHDR->code == TCN_SELCHANGE)
	{
		// Remember what the last page was
		m_nPage = GetTabControl()->GetCurSel();
	}
	return CPropertySheet::OnNotify(wParam, lParam, pResult);
}
